"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const rehype_format_1 = tslib_1.__importDefault(require("rehype-format"));
const chalk_1 = tslib_1.__importDefault(require("chalk"));
const path_1 = tslib_1.__importDefault(require("path"));
const glob_1 = tslib_1.__importDefault(require("glob"));
const unified_1 = tslib_1.__importDefault(require("unified"));
const fs_extra_1 = tslib_1.__importDefault(require("fs-extra"));
const rehype_raw_1 = tslib_1.__importDefault(require("rehype-raw"));
const gray_matter_1 = tslib_1.__importDefault(require("gray-matter"));
const remark_parse_1 = tslib_1.__importDefault(require("remark-parse"));
const remark_rehype_1 = tslib_1.__importDefault(require("remark-rehype"));
const rehype_stringify_1 = tslib_1.__importDefault(require("rehype-stringify"));
const rehype_slug_1 = tslib_1.__importDefault(require("rehype-slug"));
const fs_extra_promise_1 = tslib_1.__importDefault(require("fs-extra-promise"));
const remark_frontmatter_1 = tslib_1.__importDefault(require("remark-frontmatter"));
const engine_1 = require("./engine");
const generator_1 = require("./generator");
const graphql_1 = require("./graphql");
const edge_1 = require("./edge");
const defaultConfig = {
    sourcePath: "src",
    publicPath: "public",
    basePath: "/",
    staticPath: "static",
    assetsPath: "assets",
    graphQlPath: "graphql"
};
exports.compiler = (options = defaultConfig) => tslib_1.__awaiter(this, void 0, void 0, function* () {
    const isTestRunner = process.env.NODE_ENV === "test";
    const isProduction = process.env.NODE_ENV === "production";
    const isDevelopment = !isProduction && !isTestRunner;
    const startTime = process.hrtime();
    const basePath = options.basePath;
    const sourcePath = path_1.default.resolve(basePath, options.sourcePath);
    const publicPath = path_1.default.resolve(basePath, options.publicPath);
    const staticPath = path_1.default.resolve(sourcePath, options.staticPath);
    const graphQlPath = path_1.default.resolve(sourcePath, options.graphQlPath);
    const staticDestinationPath = path_1.default.resolve(publicPath, options.staticPath);
    const pagesPath = `${sourcePath}/pages`;
    function writeFile(destination, content) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield fs_extra_1.default.ensureFile(destination);
            yield fs_extra_promise_1.default.writeFileAsync(destination, content, { encoding: "utf8" });
        });
    }
    // tslint:disable-next-line no-unused-expression
    !isTestRunner && console.log(chalk_1.default.green("Building site..."));
    const engine = new engine_1.HandlebarsEngine({
        partialsDir: `${sourcePath}/partials`,
        layoutsDir: `${sourcePath}/layouts`,
        helpersDir: `${sourcePath}/helpers`
    });
    const generator = new generator_1.Generator(engine);
    // clean output directory
    const deletionList = glob_1.default.sync(`!(${options.staticPath}|${options.assetsPath})**`, { cwd: publicPath });
    for (const d of deletionList) {
        const p = path_1.default.resolve(publicPath, d);
        const stats = fs_extra_1.default.statSync(p);
        if (stats.isFile()) {
            yield fs_extra_promise_1.default.removeAsync(p);
        }
        if (stats.isDirectory()) {
            yield fs_extra_promise_1.default.emptyDir(p);
        }
    }
    // Collect a list of files
    const files = glob_1.default.sync("**/*.@(md|markdown)", { cwd: pagesPath });
    const pages = [];
    for (const f of files) {
        // raw content of markdown source
        const fContent = yield fs_extra_promise_1.default.readFileAsync(`${pagesPath}/${f}`, {
            encoding: "utf8"
        });
        // process markdown file from source folder
        const result = yield unified_1.default()
            .use(remark_parse_1.default)
            .use(remark_frontmatter_1.default, ["yaml", "toml"])
            .use(remark_rehype_1.default, { allowDangerousHTML: true })
            .use(rehype_slug_1.default)
            .use(rehype_raw_1.default)
            .use(rehype_format_1.default)
            .use(rehype_stringify_1.default)
            .process(fContent);
        // extract html content
        const { content } = gray_matter_1.default(String(result), {
            excerpt: true
        });
        // extract markdown formatted excerpt and frontmatter
        const { data, excerpt: MarkdownExcerpt = "" } = gray_matter_1.default(String(fContent), {
            excerpt: true
        });
        // convert markdown excerpt to html
        const convertedExcerpt = yield unified_1.default()
            .use(remark_parse_1.default)
            .use(remark_rehype_1.default, { allowDangerousHTML: true })
            .use(rehype_raw_1.default)
            .use(rehype_format_1.default)
            .use(rehype_stringify_1.default)
            .process(Buffer.from(MarkdownExcerpt));
        // extract markdown formatted from unified object
        const { excerpt } = gray_matter_1.default(String(convertedExcerpt), {
            excerpt: true
        });
        // get dirname and filename from original file
        const { dir, name } = path_1.default.parse(f);
        // choose edge name
        const fileName = name;
        // create Edge with collected information
        const page = new edge_1.Edge(fileName, content, // html content
        data, // frontmatter
        publicPath, // stasis public path (used to generate relative path)
        `${publicPath}${dir && "/" + dir}`, // destination path
        excerpt // html excerpt
        );
        pages.push(page);
    }
    const { createRoot, createSchema } = require(path_1.default.relative(__dirname, graphQlPath));
    for (const p of pages) {
        // Serialize pages
        const PageObjects = pages.map((page) => {
            const isActive = page.getDestinationPath() === p.getDestinationPath();
            return page.asObject(isActive);
        });
        let result = {};
        if (p.hasQuery()) {
            // create graphql schema and root
            const root = createRoot(PageObjects, options);
            const schema = createSchema(PageObjects, options);
            // create query function
            const query = graphql_1.makeGraph(schema, root);
            // get query result
            result = yield query(p.query);
        }
        // generate rendered html result
        const htmlOutput = yield generator.render({
            body: p.html,
            frontmatter: p.frontmatter,
            title: p.frontmatter.title || p.fileName,
            slug: p.slug,
            excerpt: p.excerpt,
            query: result,
            isProduction,
            isDevelopment
        });
        // write result to disk
        yield writeFile(`${p.getDestinationPath()}`, htmlOutput);
    }
    const rootFiles = glob_1.default.sync(`*.*`, { cwd: sourcePath });
    // Copy files from the root of the `src` directory to the root of the output directory
    for (const rootFile of rootFiles) {
        yield fs_extra_promise_1.default.copyAsync(sourcePath + "/" + rootFile, publicPath + "/" + rootFile, {
            overwrite: true,
            preserveTimestamps: true
        });
    }
    // copy static asset folder
    yield fs_extra_promise_1.default.copyAsync(staticPath, staticDestinationPath, {
        overwrite: true,
        preserveTimestamps: true
    });
    // display build time
    const timeDiff = process.hrtime(startTime);
    const duration = timeDiff[0] * 1000 + timeDiff[1] / 1e6;
    // tslint:disable-next-line no-unused-expression
    !isTestRunner &&
        console.log(chalk_1.default.green(`Site built successfully in ${duration}ms`));
    return;
});
